/***************************************************************************
                          main.c  -  description
                             -------------------
    begin                : Mon Nov 12 06:40:29 EST 2001
    copyright            : (C) 2001 by Matt Sabins
    email                : sabinsmh@clarkson.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <regex.h>
#include <string.h>

#include "gc.h"
#include "parsertypes.h"

#define MATCH_SIZE 32
#define ERROR_BUF_SIZE 255

/*Regex implementation details.*/
regmatch_t pmatch[MATCH_SIZE];
char error_buf[255];
char str_buf[255];

/*This blocks gets the file name should be the first line*/
char* get_file_name(char* str) {
	return gcsubstr(str,8,strlen(str)-2);
}

char* file_name;

/* This block of code will get the source line number.
        .stabn 68,0,(lineno),.LM(internal label)-(function_name)
*/

char stabn_str1 [] = "\t\\.stabn 68,0,[0-9]+,\\.LM[0-9]+-";
regex_t stabn;

int is_lineno(char* str) {
	return REG_NOMATCH != regexec(&stabn,str,MATCH_SIZE,pmatch,0);
}

int get_lineno(char* str) {
	char* temp;
	int i = 13,result;
	for(;str[i] != ',';i++);
	temp = gcsubstr(str,13,i);
	result = atoi(temp);
	return result;
}

/*end of line getting code*/


/* This block of code is used to detect the start and the end of a function*/
char* fnct_name = NULL;

char fnct_start_str [] = "\t\\.type\t .*@function\n";
regex_t fnct_start;

char fnct_end_str1 [] = "\t\\.size\t ";
char fnct_end_str2 [] = ",\\.Lfe.*-";

regex_t fnct_end;

int is_in_function = 0;

char* get_function_name(char* str) {
	return gcsubstr(str,8,strlen(str) - 11);
}

int is_function_start(char* str) {
	char *stabn_str;
	int result;
	size_t size;
	
	
	if(REG_NOMATCH == regexec(&fnct_start,str,MATCH_SIZE,pmatch,0))
		return 0;
	
	/*
		Compute the regular expression for the next function.
	*/	
	is_in_function = 1;
	
	if(fnct_name)
		free(fnct_name);
	
	fnct_name = get_function_name(str);
	
	sprintf(str_buf,"%s%s%s%s",fnct_end_str1,fnct_name,fnct_end_str2,fnct_name);
			
	result = regcomp(&fnct_end,str_buf,REG_EXTENDED);
	if(result != 0) {
		fputs("Problem",stderr);
		size = regerror(result,&fnct_end,error_buf,ERROR_BUF_SIZE);
		fprintf(stderr,"%d: %s",size,error_buf);
		return result;
	}
	
	/*
		Compute the regular expression for .stabn lineno.
	*/
	stabn_str = gcstrcat(stabn_str1,fnct_name);

	result = regcomp(&stabn,stabn_str,REG_EXTENDED);
	if(result != 0) {
		fputs("Problem",stderr);
		size = regerror(result,&stabn,error_buf,ERROR_BUF_SIZE);
		fprintf(stderr,"%d: %s",size,error_buf);
		return result;
	}
	return 1;
}

int is_function_end(char* str) {
	if(is_in_function && REG_NOMATCH != regexec(&fnct_end,str,MATCH_SIZE,pmatch,0)) {
		is_in_function = 0;
		fnct_name = NULL;
		return 1;
	}
	return 0;
}

/*end start end code*/

/*Need code to keep track of local string declarations.
	So I can add string declarations in with it.
	I also need to be able to generate the next ones.
*/

int last_LC_no = 1;
char last_LC_buf[32];
char LC_label_str [] = "\\.LC[0-9]+:\n";
regex_t LC_label;

void update_LC_label(char* str) {
	if(REG_NOMATCH == regexec(&LC_label,str,MATCH_SIZE,pmatch,0))
		return ;
		
	last_LC_no++;
}

char* generate_next_LC() {
	sprintf(last_LC_buf,".LC%d",++last_LC_no);
	return last_LC_buf;
}


int last_localjump_no = 1;
char last_localjump_buf[32];
char localjump_label_str [] = "\\.L[0-9]+:\n";
regex_t localjump_label;

void update_localjump_label(char* str) {
	char* colon,* begin;
	int num;
	if(REG_NOMATCH == regexec(&localjump_label,str,MATCH_SIZE,pmatch,0))
		return;	
	
	strcpy(last_localjump_buf,str);
	colon = strchr(last_localjump_buf,':');
	*colon = '\0';
	begin = last_localjump_buf + 2;
	num = atoi(begin);
	if(num > last_localjump_no) {
		last_localjump_no = num;
	}
}

char* generate_next_localjump() {
	sprintf(last_localjump_buf,".L%d",++last_localjump_no);
	return last_localjump_buf;
}


int compile_regex() {
	int result;
	size_t size;
	result = regcomp(&fnct_start,fnct_start_str,REG_EXTENDED);
	if(result != 0) {
		fputs("Problem",stderr);
		size = regerror(result,&fnct_start,error_buf,ERROR_BUF_SIZE);
		fprintf(stderr,"%d: %s",size,error_buf);
		return result;
	}
	result = regcomp(&LC_label,LC_label_str,REG_EXTENDED);
	if(result != 0) {
		fputs("Problem",stderr);
		size = regerror(result,&LC_label,error_buf,ERROR_BUF_SIZE);
		fprintf(stderr,"%d: %s",size,error_buf);
		return result;
	}	
	result = regcomp(&localjump_label,localjump_label_str,REG_EXTENDED);
	if(result != 0) {
		fputs("Problem",stderr);
		size = regerror(result,&localjump_label,error_buf,ERROR_BUF_SIZE);
		fprintf(stderr,"%d: %s",size,error_buf);
		return result;
	}		
	
	return 0;
}

void destroy_regex() {
	regfree(&fnct_start);
}


/*
			int regcomp(regex_t *preg, const char *regex, int cflags);
       int regexec(const   regex_t  *preg,  const  char  *string,
                   size_t  nmatch,   regmatch_t   pmatch[],   int
                   eflags);
       size_t regerror(int  errcode,  const  regex_t  *preg, char
                       *errbuf, size_t errbuf_size);
       void regfree(regex_t *preg);

*/

int main(int argc, char *argv[])
{
	FILE* input,*output;
	char str[255];
	memaddr_node_t* node = NULL,* cur = NULL,*head = NULL;
	int lineno = 1;
	
	if(argc == 2) {
		input = fopen(argv[1],"r");
		if(input == NULL) {
			fprintf(stderr,"Error opening %s\n",argv[1]);
		}
	}else {
			fprintf(stderr,"Arg: assembly code %s\n",argv[1]);
			return -1;	
	}
	
	output = fopen("lockout.s","w");

	if(compile_regex()) {
		fputs("Compilation Error",stderr);
		return -1;
	}
		
	if(startgc()) {
		fprintf(stderr,"Memory Error.\n");
		return -1;
	}
	
	if(!fgets(str,255,input)) {
		return -1;
	}
	
	while(fgets(str,255,input)) {
		lineno++;
		if(!is_in_function) {
			update_LC_label(str);
			if(is_function_start(str))
				continue;
		}
		else {
			update_localjump_label(str);	
			if(is_function_end(str)) {
				continue;
			}
			else if(is_lineno(str)) {
				get_lineno(str);
			}
			else if(NULL != (node = parse_mnemonic(str))) {  //make the address to check list.
				node->lineno = lineno;
				if(head == NULL) {
					cur = head = node;
				}else {
					cur->next = node;
					cur = node;
				}
			}
			else if(NULL != (node = parse_mutex_lock(str))) {  //make the address to check list.
				node->lineno = lineno;
				if(head == NULL) {
					cur = head = node;
				}else {
					cur->next = node;
					cur = node;
				}
			}
		}
	}
	
	rewind(input);
	
	lineno = 0;
	cur = head;
	
	while(fgets(str,255,input)) {
		lineno++;
		if(cur && cur->lineno == lineno) {
		
			output_assembly(output,cur,".LC0");
																	  //When outputting be sure to check to see
																		//if the access is bigger than %esp!!!!									
			cur = cur->next;
		}
		fprintf(output,"%s",str);
	}
			
	destroy_regex();
  return 0;
}
