%{
/*
	The mnemonic arguments are not SLR if it was I wouldn't be using lex and yacc.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gc.h"
#include "parsertypes.h"
#include "parseinstr.h"

int yyerror(char* s);
extern int yylex();
extern int yy_scan_string(char*);


memaddr_node_t addr;
char strbuf[255];
int strindex = 0;

char* reg;
char* mnem;
int curpos;

char* lsegment;
char* ldisp;
char* lbase;
char* lindex;
char* lscale;


%}

%union
{
	char    *sval;
}

%token <sval> IDENT NUM REG SEG
%token RPAREN LPAREN COMMA IMMED DOLLAR MINUS

%type <sval> number

%%

mnemonic: IDENT arguments {free($1);}
				;

arguments : memaddr
					| arguments COMMA memaddr
					;
					
memaddr :
				DOLLAR number { free($2); curpos++;}
				| DOLLAR IDENT { free($2); curpos++;}
				|	REG {free($1); curpos++;}
				| {strindex = 0;strbuf[0] = '\0';addr.is_valid = true;}
						disp reference {addr.position = curpos;}
				;
				
disp:
		| SEG offset {strcat(strbuf,$1);lsegment = $1;}
		| offset
		;
		
offset:
			IDENT {strcat(strbuf,$1);addr.is_static = true; ldisp = $1;}
		| number {strcat(strbuf,$1);ldisp = $1;}
		;
		
reference:
			| LPAREN {strcat(strbuf,"(");} base index scale RPAREN {strcat(strbuf,")");}
			;
			
base:
		| REG {
			if(!strcmp($1,"%esp") || !strcmp($1,"%ebp")) {
				addr.is_valid = false;
			}else {
				strcat(strbuf,$1);
			}
			lbase = $1;
		}
		;
			
index:
			| COMMA REG {
				if(!strcmp($2,"%esp") || !strcmp($2,"%ebp")) {
					addr.is_valid = false;
				}else {
					strcat(strbuf,",");
					strcat(strbuf,$2);	
				}
				lindex = $2;
			}
			;
			
scale:
			| COMMA number {
					strcat(strbuf,",");
					strcat(strbuf,$2);	
			}
		;
		
number:
			MINUS NUM {$$ = gcstrcat("-",$2);free($2);}
			| NUM {;}
		
%%

int yyerror (char* s)  /* Called by yyparse on error */
{
  printf ("%s\n", s);
  return 0;
}

memaddr_node_t* parse_mnemonic(char* line) {
	bool s_w = false,s_r = false,d_rw = false,d_w = false,d_r = false;
	mnem = gcword(line,0);
	
	memset(&addr,0,sizeof(memaddr_node_t));
	curpos = 1;
	
	if(is_dbl_rw_instr(mnem) )
		d_rw = true;
	else if(is_dbl_ronly_instr(mnem))
		d_r = true;		
	else if(is_dbl_wonly_instr(mnem))
		d_w = true;	
	else if(is_sngl_write_instr(mnem))
		s_w = true;
	else if(is_sngl_read_instr(mnem))
		s_r = true;
	else {
		free(mnem);
		mnem = NULL;
		return NULL;
	}

	free(mnem);		
	mnem = NULL;

	lsegment = NULL;
	ldisp = NULL;
	lbase = NULL;
	lindex = NULL;
	lscale = NULL;

		
	yy_scan_string(line);

	yyparse();
	
	if(!addr.is_valid) {
	
		if(lsegment)
			free(lsegment);
		if(ldisp)
			free(ldisp);
		if(lsegment)
			free(lbase);
		if(lbase)
			free(lindex);
		if(lsegment)
			free(lindex);
		return NULL;	
	}
		
	if(d_rw) {
		if(addr.position == 2 && addr.is_static) {
			addr.line_type = L_STATIC_WRITE;
		}
		else if(addr.position == 1 && addr.is_static) {
			addr.line_type = L_STATIC_READ;
		}
		else if(addr.position == 2 && !addr.is_static) {
			addr.line_type = L_DYNAMIC_WRITE;
		}
		else if(addr.position == 1 && !addr.is_static) {
			addr.line_type = L_DYNAMIC_READ;
		}		
	}
	else if(d_r || s_r) {
		if(addr.is_static) {
			addr.line_type = L_STATIC_READ;
		}	
		else if(!addr.is_static) {
			addr.line_type = L_DYNAMIC_READ;
		}			
	}
	else if(d_w || s_w) {
		if(addr.is_static) {
	 		addr.line_type = L_STATIC_WRITE;
		}	
		else if(!addr.is_static) {
			addr.line_type = L_DYNAMIC_WRITE;
		}			
	}
	
	addr.address = gcdup(strbuf);
	addr.lock_type = L_NOLOCK;
	return gccopy(&addr,sizeof(memaddr_node_t));
}
